/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

/** Tests for the Max-Sum algorithm */
package ch.epfl.lia.frodo.algorithms.maxsum.tests;

import java.util.List;

import org.jdom.Document;
import org.jdom.Element;

import ch.epfl.lia.frodo.algorithms.RandGraphFactory;
import ch.epfl.lia.frodo.algorithms.RandGraphFactory.Graph;
import ch.epfl.lia.frodo.algorithms.Solution;
import ch.epfl.lia.frodo.algorithms.XCSPparser;
import ch.epfl.lia.frodo.algorithms.dpop.DPOPsolver;
import ch.epfl.lia.frodo.algorithms.dpop.count.SolutionCounter;
import ch.epfl.lia.frodo.algorithms.maxsum.MaxSum;
import ch.epfl.lia.frodo.algorithms.maxsum.MaxSumSolver;
import ch.epfl.lia.frodo.algorithms.test.AllTests;
import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.AddableInteger;
import ch.epfl.lia.frodo.solutionSpaces.AddableReal;
import junit.extensions.RepeatedTest;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/** Unit tests for the Max-Sum algorithm
 * @author Thomas Leaute
 * @param <V> the type used for variable values
 * @param <U> the type used for utility values
 */
public class MaxSumTests< V extends Addable<V>, U extends Addable<U> > extends TestCase {
	
	/** Maximum number of variables in the test problems */
	private static final int maxNbrVars = 20;
	
	/** Maximum number of edges in the test constraint graphs */
	private static final int maxNbrEdges = 190;
	
	/** Maximum number of agents in the test problems */
	private static final int maxNbrAgents = 20;
	
	/** @return the suite of unit tests */
	public static TestSuite suite () {
		
		TestSuite suite = new TestSuite ("Tests for Max-Sum");
		
		TestSuite tmp = new TestSuite ("Pure acyclic maximization problems");
		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableInteger> (true, AddableInteger.class, 0.0, false, true), 1000));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Pure acyclic minimization problems");
		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableInteger> (false, AddableInteger.class, 0.0, false, true), 1000));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Pure maximization problems");
		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableInteger> (true, AddableInteger.class, 0.0, false, false), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Pure minimization problems");
		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableInteger> (false, AddableInteger.class, 0.0, false, false), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Pure minimization problems with real-valued costs");
		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableReal> (false, AddableReal.class, 0.0, false, false), 100));
		suite.addTest(tmp);
		
		/// @bug Max-Sum may not terminate when using TCP pipes because there is not idleness detection
//		tmp = new TestSuite ("Pure minimization problems with TCP pipes");
//		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableInteger> (false, AddableInteger.class, 0.0, true, false), 100));
//		suite.addTest(tmp);
		
		tmp = new TestSuite ("Maximization problems");
		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableInteger> (true, AddableInteger.class, AllTests.DEFAULT_P2, false, false), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Minimization problems");
		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableInteger> (false, AddableInteger.class, AllTests.DEFAULT_P2, false, false), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Minimization problems with real-valued costs");
		tmp.addTest(new RepeatedTest (new MaxSumTests<AddableInteger, AddableReal> (false, AddableReal.class, AllTests.DEFAULT_P2, false, false), 100));
		suite.addTest(tmp);
		
		return suite;
	}
	
	/** Whether to maximize utility or minimize cost */
	private boolean maximize;
	
	/** The class of U */
	private Class<U> classOfU;
	
	/** The constraint tightness */
	private final double p2;
	
	/** Whether to use TCP pipes */
	private final boolean useTCP;
	
	/** Whether the graph should be acyclic */
	private final boolean acyclic;
	
	/** Constructor
	 * @param maximize 	Whether to maximize utility or minimize cost
	 * @param classOfU 	The class of U
	 * @param p2 		The constraint tightness
	 * @param useTCP 	Whether to use TCP pipes
	 * @param acyclic 	Whether the graph should be acyclic
	 */
	public MaxSumTests (boolean maximize, Class<U> classOfU, double p2, boolean useTCP, boolean acyclic) {
		super ("test");
		this.maximize = maximize;
		this.classOfU = classOfU;
		this.p2 = p2;
		this.useTCP = useTCP;
		this.acyclic = acyclic;
	}
	
	/** The test method 
	 * @throws Exception if an error occurs
	 */
	@SuppressWarnings("unchecked")
	public void test () throws Exception {
		
		// Construct a random problem instance
		Document problem;
		if (acyclic) {
			Graph graph = RandGraphFactory.getAcyclicGraph(10, 3);
			problem = AllTests.generateProblem(graph, maximize, 0, false, p2);
		} else 
			problem = AllTests.createRandProblem(maxNbrVars, maxNbrEdges, maxNbrAgents, this.maximize, this.p2);
		
		// Get the agent configuration file
		Document agentConfig = XCSPparser.parse("src/ch/epfl/lia/frodo/algorithms/maxsum/MaxSumAgent.xml", false);
		for (Element module : (List<Element>) agentConfig.getRootElement().getChild("modules").getChildren()) {
			if (module.getAttributeValue("className").equals(MaxSum.class.getName())) {
				module.setAttribute("rescale", Boolean.toString(this.p2 == 0.0));
				if (acyclic) 
					module.setAttribute("nbrCycles", "50000");
			}
		}
		
		// Instantiate the solver
		MaxSumSolver<V, U> solver = new MaxSumSolver<V, U> (agentConfig, this.useTCP);
		solver.setUtilClass(this.classOfU);
		
		// Solve the problem
		Solution<V, U> sol = solver.solve(problem);
		assertTrue ("Max-Sum timed out", sol != null);
		
		// Check that the declared utility corresponds to the solution found
		XCSPparser<V, U> parser = new XCSPparser<V, U > (problem);
		parser.setUtilClass(this.classOfU);
		assertEquals (parser.getUtility(sol.getAssignments()).getUtility(0), sol.getUtility());
		
		// Skip the completeness test if the problem is not acyclic, or has more than one optimal solutions
		if (this.acyclic && new SolutionCounter<AddableInteger, AddableInteger>().count(problem) <= 1) 
			assertEquals(new DPOPsolver<AddableInteger, AddableInteger>().solve(problem).getUtility(), sol.getUtility());
	}

}
